require "Window"
require "GameLib"
require "ChatSystemLib"
require "AbilityBook"
require "MacrosLib"
require "GuildLib"
require "GroupLib"
require "MatchingGame" -- Matchmaking for group finder, used to find Roles available
require "XmlDoc"
require "CharacterTitle"
local SuitUp = {}

--region Constants
-----------------------------------------------------------------------------------------------
-- Constants
local kSuitUp =
{
    Name = "Suit Up!",
}
local ArrangeVert = {FromTop = 0, FromCenter = 1, FromBottom = 2}
--endregion

--region Addon Initialization and State
----------------------------------------------------------------------------------------------
-- Initialization
-----------------------------------------------------------------------------------------------
function SuitUp:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self
    o.tOptions = 
    {
        Closet = {}, 
        SuitsButtonPos = nil,

        ShowQuickButton         = true; 
        HideQuickButtonInCombat = false;
        LockQuickButton         = false;
        AutoCloseSuitList       = true; 
        ReportAllChanges        = false;
        DisableChat             = false;
        ConfirmSuitDeletes      = true; 
        ConfirmActionDeletes    = false;
        UseChatFallback         = true;
    }
    return o
end

function SuitUp:Init()
    local bHasConfigureFunction = true
    local strConfigureButtonText = kSuitUp.Name
    local tDependencies = {"Xowl:LibCombo-1.0",}

    -- Set up some working state
    self.Editor                 = {}
    self.Editor.CurrentSuit     = nil
    self.Editor.CurrentArticle  = nil

    Apollo.RegisterAddon(self, bHasConfigureFunction, strConfigureButtonText, tDependencies)
end

function SuitUp:MakeInitialData()
    local TestCloset =
    {
        {Name = "First",    Articles = {{ArticleType=self.SuitCatalog.eArticleType.Costume, ArticleVersion=1, CostumeId=1,},
                                        {ArticleType=self.SuitCatalog.eArticleType.Say,     ArticleVersion=1, Message="Changing to DPS spec, please give me a second", Channel=4, Percent=1.0,},
                                        {ArticleType=self.SuitCatalog.eArticleType.Spec,    ArticleVersion=1, SpecId=1,},
                                       },
        },
        {Name = "Second",   Articles = {{ArticleType=self.SuitCatalog.eArticleType.Costume, ArticleVersion=1, CostumeId=3,},
                                        {ArticleType=self.SuitCatalog.eArticleType.Say,     ArticleVersion=1, Message="Changing to Heal spec, please give me a second", Channel=4, Percent=1.0,},
                                        {ArticleType=self.SuitCatalog.eArticleType.Spec,    ArticleVersion=1, SpecId=2,},
                                       },
        },
    }

    -- Fix up some test data. Take the "data" portion from TestCloset, which should be like what we persist, and add the methods to them
    for _,v in ipairs(TestCloset) do
        local s = {Name = v.Name, Articles={}}
        
        for _,art in ipairs(v.Articles) do 
            table.insert(s.Articles, self.SuitCatalog.ArticlePatterns:FromData(art)); 
        end

        self.tOptions.Closet[#self.tOptions.Closet + 1] = s
    end
end

function SuitUp:OnSave(eLevel)
    if eLevel ~= GameLib.CodeEnumAddonSaveLevel.Character then return; end
    
    local data = {Closet={}}

    -- Save window loactions
    if self.wndSuits then
        local nLeft, nTop, nRight, nBottom = self.wndSuits:GetAnchorOffsets()
        data.SuitsButtonPos = {Left = nLeft, Top = nTop, Right = nRight, Bottom = nBottom}
    end

    if self.wndConfig then
        local nLeft, nTop, nRight, nBottom = self.wndConfig:GetAnchorOffsets()
        data.ConfigPos = {Left = nLeft, Top = nTop, Right = nRight, Bottom = nBottom}
    end

    -- Save general options
    data.ShowQuickButton         = self.tOptions.ShowQuickButton      
    data.HideQuickButtonInCombat = self.tOptions.HideQuickButtonInCombat
    data.LockQuickButton         = self.tOptions.LockQuickButton      
    data.ReportAllChanges        = self.tOptions.ReportAllChanges     
    data.DisableChat             = self.tOptions.DisableChat          
    data.ConfirmSuitDeletes      = self.tOptions.ConfirmSuitDeletes       
    data.ConfirmActionDeletes    = self.tOptions.ConfirmActionDeletes       
    data.AutoCloseSuitList       = self.tOptions.AutoCloseSuitList
    data.UseChatFallback         = self.tOptions.UseChatFallback

    -- Save the closet data
    for is,s in ipairs(self.tOptions.Closet) do
        local newsuit = {Name = s.Name, Articles={}}
        for ia,a in ipairs(s.Articles) do
            newsuit.Articles[#newsuit.Articles + 1] = a:ToData()
        end
        data.Closet[#data.Closet + 1] = newsuit
    end

    return {SaveVersion = 1, SaveData = data} 
end

function SuitUp:OnRestore(eLevel, tData)
    if eLevel == GameLib.CodeEnumAddonSaveLevel.Character and tData ~= nil then
        
        -- Save the data in case we need it later (like when files have finished loading)
        self.RestoredData = tData 

        if self.RestoredData.SaveVersion == 1 then
            -- Button positions
            local SavedButtonPos = self.RestoredData.SaveData.SuitsButtonPos

            if SavedButtonPos ~= nil then
                self.tOptions.SuitsButtonPos = {Left = SavedButtonPos.Left, Right = SavedButtonPos.Right, Top = SavedButtonPos.Top, Bottom = SavedButtonPos.Bottom}
                if self.wndSuits then self.wndSuits:SetAnchorOffsets(self.tOptions.SuitsButtonPos.Left, self.tOptions.SuitsButtonPos.Top, self.tOptions.SuitsButtonPos.Right, self.tOptions.SuitsButtonPos.Bottom); end
            end

            local ConfigPos = self.RestoredData.SaveData.ConfigPos
            if ConfigPos ~= nil then
                self.tOptions.ConfigPos = {Left = ConfigPos.Left, Right = ConfigPos.Right, Top = ConfigPos.Top, Bottom = ConfigPos.Bottom}
                if self.wndConfig then self.wndConfig:SetAnchorOffsets(self.tOptions.ConfigPos.Left, self.tOptions.ConfigPos.Top, self.tOptions.ConfigPos.Right, self.tOptions.ConfigPos.Bottom); end
            end
            
            -- General options
            self.tOptions.ShowQuickButton         = self.RestoredData.SaveData.ShowQuickButton      
            self.tOptions.HideQuickButtonInCombat = self.RestoredData.SaveData.HideQuickButtonInCombat
            self.tOptions.LockQuickButton         = self.RestoredData.SaveData.LockQuickButton      
            self.tOptions.AutoCloseSuitList       = self.RestoredData.SaveData.AutoCloseSuitList
            self.tOptions.ReportAllChanges        = self.RestoredData.SaveData.ReportAllChanges     
            self.tOptions.DisableChat             = self.RestoredData.SaveData.DisableChat          
            self.tOptions.ConfirmSuitDeletes      = self.RestoredData.SaveData.ConfirmSuitDeletes
            self.tOptions.ConfirmActionDeletes    = self.RestoredData.SaveData.ConfirmActionDeletes
            self.tOptions.UseChatFallback         = self.RestoredData.SaveData.UseChatFallback

            -- Any options that were nil should get default values
            if nil == self.tOptions.ShowQuickButton         then self.tOptions.ShowQuickButton         = true;  end
            if nil == self.tOptions.HideQuickButtonInCombat then self.tOptions.HideQuickButtonInCombat = false; end
            if nil == self.tOptions.LockQuickButton         then self.tOptions.LockQuickButton         = false; end
            if nil == self.tOptions.AutoCloseSuitList       then self.tOptions.AutoCloseSuitList       = true;  end
            if nil == self.tOptions.ReportAllChanges        then self.tOptions.ReportAllChanges        = false; end
            if nil == self.tOptions.DisableChat             then self.tOptions.DisableChat             = false; end
            if nil == self.tOptions.UseChatFallback         then self.tOptions.UseChatFallback         = true;  end
            if nil == self.tOptions.ConfirmSuitDeletes      then self.tOptions.ConfirmSuitDeletes      = true;  end
            if nil == self.tOptions.ConfirmActionDeletes    then self.tOptions.ConfirmActionDeletes    = false; end

            -- Saved suits (the big, important data!)
            self:LoadSavedCloset()

            -- Set any option-related properties we can on windows and so forth
            self:ApplyOptions()
        end
    end
end

-- Tries to read the data from Restore, if we have it
function SuitUp:LoadSavedCloset()
    if self.SuitCatalog ~= nil and self.RestoredData ~= nil and self.RestoredData.SaveData ~= nil then
        self.tOptions.Closet = {}
        for is,s in ipairs(self.RestoredData.SaveData.Closet) do
            local newsuit = {Name = s.Name, Articles = {}}
            for ia,a in ipairs(s.Articles) do
                newsuit.Articles[#newsuit.Articles + 1] = self.SuitCatalog.ArticlePatterns:FromData(a)
            end
            self.tOptions.Closet[#self.tOptions.Closet + 1] = newsuit
        end
    end
end

-----------------------------------------------------------------------------------------------
function SuitUp:OnLoad()
    self.xmlDoc = XmlDoc.CreateFromFile("SuitUp.xml")
    self.xmlDoc:RegisterCallback("OnFormXMLLoaded", self)
end

-----------------------------------------------------------------------------------------------
function SuitUp:OnFormXMLLoaded()

    if self.xmlDoc ~= nil and self.xmlDoc:IsLoaded() then
        -------------------------------------------------------------------------------------------
        -- Execute the Suits file to get the data structures and constants we need
        -- I'm using the dreaded long bracket string constant form for an actual string constant
        -- here because I can't decide how much I hate it and forcing myself to look at it will
        -- help me figure that out.
        self.SuitsFileName = Apollo.GetAssetFolder() .. [[\]] .. "SuitPatterns.lua"
        self.SuitCatalog = dofile(self.SuitsFileName)

        self:LoadSavedCloset() -- Now that we have the patterns, process the saved data we got in :OnRestore

        -------------------------------------------------------------------------------------------
        -- Load and set up the config/editor form. This is shown from the Config button on the        -- suits popup, the /suitup command, or the SuitUp option on the system escape menu
        self.wndConfig = Apollo.LoadForm(self.xmlDoc, "SuitUpConfigForm", nil, self)
        if self.wndConfig == nil then Apollo.AddAddonErrorText(self, "Could not load the configuration form.") return; end
        if self.tOptions.ConfigPos ~= nil then self.wndConfig:SetAnchorOffsets(self.tOptions.ConfigPos.Left, self.tOptions.ConfigPos.Top, self.tOptions.ConfigPos.Right, self.tOptions.ConfigPos.Bottom); end
        self.wndConfig:Show(false, true)

        local wndConfigContent = self.wndConfig:FindChild("MainFrame:ContentsFrame")
        local tabSuitEditor    = self.wndConfig:FindChild("MainFrame:TabFrame:SuitsEditorTab")
        local tabOptions       = self.wndConfig:FindChild("MainFrame:TabFrame:SuitUpOptionsTab")
        local tabHelp          = self.wndConfig:FindChild("MainFrame:TabFrame:SuitUpHelpTab")
        local wndOptions       = Apollo.LoadForm(self.xmlDoc, "ConfigOptionsContent", wndConfigContent, self)
        local wndSuitEditor    = Apollo.LoadForm(self.xmlDoc, "SuitsEditorContent", wndConfigContent, self)
        local wndHelp          = Apollo.LoadForm(self.xmlDoc, "HelpContent", wndConfigContent, self)

        -- These buttons are a radio group. Attaching a window to them makes them toggle the attached window visibility. Presto: tabs.
        tabSuitEditor:AttachWindow(wndSuitEditor);  tabSuitEditor:SetCheck(true)
        tabOptions:AttachWindow(wndOptions);        tabOptions:SetCheck(false)
        tabHelp:AttachWindow(wndHelp);              tabHelp:SetCheck(false)

        self.HelpFileName = Apollo.GetAssetFolder() .. [[\]] .. "SuitUpHelp.lua"
        local HelpAML = dofile(self.HelpFileName)
        wndHelp:FindChild("MLContent"):SetAML(HelpAML)
        wndHelp:FindChild("MLContent"):SetVScrollPos(0)

        -------------------------------------------------------------------------------------------
        -- Load and set up the Suits popup (the "closet"). This is the onscreen button that users        
        -- interact with to change suits (to "Suit Up!")
        self.wndSuits = Apollo.LoadForm(self.xmlDoc, "SuitsForm", nil, self)
        if self.wndSuits == nil then Apollo.AddAddonErrorText(self, "Could not load the Suits form.") return; end
        if self.tOptions.SuitsButtonPos ~= nil then self.wndSuits:SetAnchorOffsets(self.tOptions.SuitsButtonPos.Left, self.tOptions.SuitsButtonPos.Top, self.tOptions.SuitsButtonPos.Right, self.tOptions.SuitsButtonPos.Bottom); end
        self.wndSuitsPopout = self.wndSuits:FindChild("SuitsPopupList")
        
        -- Magic... attach a window to a button that's marked as a checkbox, and the button toggles the window visibility
        self.wndSuits:FindChild("SuitsPopoutBtn"):AttachWindow(self.wndSuitsPopout)
        self:RebuildCloset()
        self.wndSuits:Show(true)


        -- Register handlers for events, slash commands and timer, etc.
        Apollo.RegisterEventHandler("UnitEnteredCombat", "OnEnteredCombat", self)
        Apollo.RegisterSlashCommand("suitup", "OnSlashSuitUp", self)

        -- Apply any options (window styles, etc.)
        self:ApplyOptions()
    end
end
--endregion Addon Initialization and State

--region Apollo Event Handlers
-----------------------------------------------------------------------------------------------
-- Apollo Handlers
-----------------------------------------------------------------------------------------------
-- on slash command "/suitup"
function SuitUp:OnSlashSuitUp(strCommand,strArgs,c)
    for _,v in pairs(self.tOptions.Closet) do
        if string.lower(v.Name) == string.lower(strArgs) then
            self:ApplySuit(v)
            return
        end
    end

    self.wndConfig:Invoke() -- show the window
end

function SuitUp:OnEnteredCombat(unit, b)
    if self.tOptions.HideQuickButtonInCombat and unit == GameLib.GetPlayerUnit() then
        self.wndSuits:Show(not b)
    end
end
--endregion Apollo Event Handlers

-----------------------------------------------------------------------------------------------
-- Configuration Form
-----------------------------------------------------------------------------------------------
function SuitUp:OnOK()
    self.wndConfig:Close()
end

function SuitUp:OnCancel()
    self.wndConfig:Close()
end

---------------------------------------------------------------------------------------------------
-- ConfigOptionsContent Functions
---------------------------------------------------------------------------------------------------
function SuitUp:OnOptionsShow(wndHandler, wndControl)
    -- OnShow is called when the window is shown because the attached checkbox state is toggled, but 
    -- not when the window is shown because the topmost parent is restored from hiding. Annoying, but
    -- shouldn't actully mess us up.
    local wnd = self.wndConfig:FindChild("ConfigOptionsContent")
    wnd:FindChild("CheckShowQuickButton")     :SetCheck(self.tOptions.ShowQuickButton)
    wnd:FindChild("CheckHideInCombat")        :SetCheck(self.tOptions.HideQuickButtonInCombat)
    wnd:FindChild("CheckLockButton")          :SetCheck(self.tOptions.LockQuickButton)
    wnd:FindChild("CheckReportSuitChange")    :SetCheck(self.tOptions.ReportAllChanges)
    wnd:FindChild("CheckDisableSay")          :SetCheck(self.tOptions.DisableChat)
    wnd:FindChild("CheckConfirmSuitDeletes")  :SetCheck(self.tOptions.ConfirmSuitDeletes)
    wnd:FindChild("CheckConfirmActionDeletes"):SetCheck(self.tOptions.ConfirmActionDeletes)
    wnd:FindChild("CheckClosePopoutOnApply")  :SetCheck(self.tOptions.AutoCloseSuitList)
    wnd:FindChild("CheckUseChatFallback")     :SetCheck(self.tOptions.UseChatFallback)
end

function SuitUp:OnOptionsChanged(wndHandler, wndControl, eMouseButton)
    local wnd = self.wndConfig:FindChild("ConfigOptionsContent")
    self.tOptions.ShowQuickButton         = wnd:FindChild("CheckShowQuickButton")     :IsChecked()
    self.tOptions.HideQuickButtonInCombat = wnd:FindChild("CheckHideInCombat")        :IsChecked()
    self.tOptions.LockQuickButton         = wnd:FindChild("CheckLockButton")          :IsChecked()
    self.tOptions.AutoCloseSuitList       = wnd:FindChild("CheckClosePopoutOnApply")  :IsChecked()
    self.tOptions.ReportAllChanges        = wnd:FindChild("CheckReportSuitChange")    :IsChecked()
    self.tOptions.DisableChat             = wnd:FindChild("CheckDisableSay")          :IsChecked()
    self.tOptions.ConfirmSuitDeletes      = wnd:FindChild("CheckConfirmSuitDeletes")  :IsChecked()
    self.tOptions.ConfirmActionDeletes    = wnd:FindChild("CheckConfirmActionDeletes"):IsChecked()
    self.tOptions.UseChatFallback         = wnd:FindChild("CheckUseChatFallback")     :IsChecked()

    self:ApplyOptions()
end

function SuitUp:OnResetWindowLocations(wndHandler, wndControl, eMouseButton)
    -- NYI
end

function SuitUp:ApplyOptions()
    if self.wndSuits then
        self.wndSuits:Show(self.tOptions.ShowQuickButton)

        if self.tOptions.HideQuickButtonInCombat and GameLib.GetPlayerUnit():IsInCombat() then 
            self.wndSuits:Show(false)
        end

        self.wndSuits:SetStyle("Moveable", not self.tOptions.LockQuickButton)

        if     self.tOptions.LockQuickButton then self.wndSuits:FindChild("BG"):SetBGOpacity(0); end
        if not self.tOptions.LockQuickButton then self.wndSuits:FindChild("BG"):SetBGOpacity(1); end
    end
end

-----------------------------------------------------------------------------------------------
-- Suits popup button internal methods
-----------------------------------------------------------------------------------------------
function SuitUp:RebuildCloset()
    -- Go through the list of suits and make a button for each in the scroll area of the popup
    -- First, clear out the old list
    local wndSuitsList = self.wndSuitsPopout:FindChild("SuitsList")
    wndSuitsList:DestroyChildren()

    for k,v in ipairs(self.tOptions.Closet) do
        -- Create a button window and parent it to the list
        local wnd = Apollo.LoadForm(self.xmlDoc, "SuitButton", wndSuitsList, self)

        -- Save the Suit entry into the button for later
        wnd:SetData(v)

        -- Now set the button's various properties
        wnd:FindChild("SuitName"):SetText(v.Name)
        wnd:FindChild("SuitIcon"):SetText(tostring(k))

        local amlTT ="<P Font=\"CRB_HeaderMedium\">".. v.Name .."</P>"
        for _,art in ipairs(v.Articles) do
            local amlTTLine = art:DescribeAML()
            if amlTTLine then amlTT = amlTT..amlTTLine; end
        end
        wnd:SetTooltip(amlTT)
    end

    -- Now some shenanigans to make the popout a reasonable size
    local nSuits = #wndSuitsList:GetChildren()
    if nSuits > 0 then
        local nMaxSuitsVisible = 7 -- Just a random max before scrolling
        local nHeight = wndSuitsList:ArrangeChildrenVert(ArrangeVert.FromTop) -- Get the height if we arrange the whole stack vertically
        local nHeightMax = (nHeight/nSuits) * nMaxSuitsVisible -- Get the height per suit times the max we want to show
        nHeight = math.min(nHeight, nHeightMax)
        local nLeft, nTop, nRight, nBottom = self.wndSuitsPopout:GetAnchorOffsets()
        -- Now, move the top to nHeight above the bottom. Bottom and Top are relative to the bottom of the screen.
        nTop = nBottom - nHeight - 98 -- The 98 is some magic padding for the sprite, which extends outside the window
        self.wndSuitsPopout:SetAnchorOffsets(nLeft, nTop, nRight, nBottom)

    else
        -- The mounts button disappears if no mounts are known, but I think we want to keep the Suits button as a way to get to config
    end
end

---------------------------------------------------------------------------------------------------
-- Suits button
-- This is the button for a specific suit in the main popup list. This is the usual way Suits are
-- accessed.
---------------------------------------------------------------------------------------------------
function SuitUp:OnSuitButton(wndHandler, wndControl)
    local suit = wndControl:GetData()
    self:ApplySuit(suit)
    if self.tOptions.AutoCloseSuitList then
        self.wndSuits:FindChild("SuitsPopoutBtn"):SetCheck(false)
    end
end

function SuitUp:ApplySuit(suit)
    if self.tOptions.ReportAllChanges then
        ChatSystemLib.PostOnChannel(ChatSystemLib.ChatChannel_Command, "SuitUp: " .. suit.Name)
    end
    for _,Article in ipairs(suit.Articles) do
        Article:Apply()
    end
end

function SuitUp:OnConfigure()
    -- Not a toggle. This is from the system config menu, so always show the config dialog
    self.wndConfig:Invoke()
end

function SuitUp:OnConfigureButton(wndHandler, wndControl)
    if self.wndConfig:IsShown() then
        self.wndConfig:Close()
        return
    end

    self.wndConfig:Invoke()
end

---------------------------------------------------------------------------------------------------
-- SuitUp Configuration Form
---------------------------------------------------------------------------------------------------
function SuitUp:OnConfigShow(wndHandler, wndControl)
    -- Ignore this if it's a bubbled event
    if wndHandler ~= wndControl or wndControl ~= self.wndConfig then return; end
    -----------------------------------------------------------------------------------------------
    -- Set up the editor tab
    local wndSuitsList = self.wndConfig:FindChild("SuitsList")
    
    -- Empty the Suits list
    wndSuitsList:DestroyChildren()

    -- Refill the Suits list
    for i,v in ipairs(self.tOptions.Closet) do
        local wnd = Apollo.LoadForm(self.xmlDoc, "SuitListEditorButton", wndSuitsList, self)
        wnd:SetData(v)
        wnd:FindChild("SuitName"):SetText(v.Name)
        wnd:FindChild("SuitIcon"):SetText(tostring(i))

        -- Set the description line to blank. Having the detailed description was too distracting.
        local amlDesc = ""
        --[[
        for _,art in ipairs(v.Articles) do
            local amlLine = art:DescribeAML()
            if amlLine then amlDesc = amlDesc..amlLine end
        end
        --]]
        local wndDesc = wnd:FindChild("SuitDescription")
        wndDesc:SetAML(amlDesc)

        -- Resize the button to match the height of its contents.
        --  Resize the Description line
        --  Get the Description line's new height
        --  Use the lower of the Description line's bottom or the Icon's bottom
        local _, nAMLTop, _, _ = wndDesc:GetAnchorOffsets()
        local _, nTextHeight = wndDesc:SetHeightToContentHeight()
        local nLeft, nTop, nRight, _ = wnd:GetAnchorOffsets()
        local _, nIconTop, _, nIconBottom = wnd:FindChild("SuitIcon"):GetAnchorOffsets()
        wnd:SetAnchorOffsets(nLeft, nTop, nRight, nTop + math.max(nAMLTop + nTextHeight + 5, nIconBottom - nIconTop))
    end
    
    -- Add the button to create a new suit
    local wndNewSuit = Apollo.LoadForm(self.xmlDoc, "NewSuitButton", wndSuitsList, self)
    --TODO: See if the client rect of the parent is a good size to scale to

    wndSuitsList:ArrangeChildrenVert(ArrangeVert.FromTop)

    -- Clear out the right Articles list
    self:ClearArticlesEditor()
end

function SuitUp:ClearArticlesEditor()
    local wndArticlesEditor = self.wndConfig:FindChild("SuitArticlesEditor")
    wndArticlesEditor:FindChild("SuitNameEdit"):SetText("")
    wndArticlesEditor:FindChild("SuitNameEdit"):Enable(false)
    wndArticlesEditor:FindChild("ArticlesList"):DestroyChildren()
    self.Editor.CurrentSuit = nil --TODO: Should ClearArticlesEditor actually remove the CurrentSuit?
    self.Editor.CurrentArticle = nil
end

function SuitUp:OnNewSuitButton(wndHandler, wndControl, eMouseButton)
    local s = {Name = "New Suit", Articles={}}
    self.tOptions.Closet[#self.tOptions.Closet + 1] = s
    self:RebuildCloset()
    self:OnConfigShow(self.wndConfig, self.wndConfig)
    self:ShowSuitInEditor(s)

    local wndSuitsList = self.wndConfig:FindChild("SuitsList")
    wndSuitsList:SetVScrollPos(wndSuitsList:GetVScrollRange())
end

function SuitUp:OnMoveSuitUp(wndHandler, wndControl, eMouseButton)
    local suit = self.Editor.CurrentSuit
    local index = nil
    for i,v in ipairs(self.tOptions.Closet) do
        if v == suit then index = i; break; end
    end
    if index == nil then Print("SuitUp: Error moving suit up in the list"); return; end

    if index == 1 then return; end -- Can't move the first item up

    self.tOptions.Closet[index], self.tOptions.Closet[index - 1] = self.tOptions.Closet[index - 1], self.tOptions.Closet[index]

    self:RebuildCloset()
    self:OnConfigShow(self.wndConfig, self.wndConfig)
    self:ShowSuitInEditor(suit)
    --TODO: Scroll to make the suit show
end

function SuitUp:OnMoveSuitDown(wndHandler, wndControl, eMouseButton)
    local suit = self.Editor.CurrentSuit
    local index = nil
    for i,v in ipairs(self.tOptions.Closet) do
        if v == suit then index = i; break; end
    end
    if index == nil then Print("SuitUp: Error moving suit down in the list"); return; end

    if index == #self.tOptions.Closet then return; end -- Can't move the first item up

    self.tOptions.Closet[index], self.tOptions.Closet[index + 1] = self.tOptions.Closet[index + 1], self.tOptions.Closet[index]

    self:RebuildCloset()
    self:OnConfigShow(self.wndConfig, self.wndConfig)
    self:ShowSuitInEditor(suit)
    --TODO: Scroll to make the suit show
end

function SuitUp:OnDeleteSuit(wndHandler, wndControl, eMouseButton)
    if not self.tOptions.ConfirmSuitDeletes then
        self:DeleteSuit(self.Editor.CurrentSuit)
        return
    end

    -- They want to confirm, so... let's confirm
    local wndConfirm = wndControl:FindChild("ConfirmDeleteSuitFrame")
    wndConfirm:FindChild("ConfirmDeleteSuitText"):SetText("Really Delete Suit '"..self.Editor.CurrentSuit.Name.."'?")
    wndControl:BringChildToTop(wndConfirm)
    self.wndConfig:BringChildToTop(wndConfirm)
    wndConfirm:Show(true)
end

function SuitUp:OnCancelDeleteSuit(wndHandler, wndControl, eMouseButton)
    -- Just shut down the display
    wndControl:GetParent():Show(false)
end

function SuitUp:OnConfirmDeleteSuit(wndHandler, wndControl, eMouseButton)
    wndControl:GetParent():Show(false)
    self:DeleteSuit(self.Editor.CurrentSuit) -- Note that this destroys the editor windows
end


function SuitUp:DeleteSuit(suit)
    local index = nil
    for i,v in ipairs(self.tOptions.Closet) do
        if v == suit then index = i; break; end
    end

    if index == nil then Print("SuitUp: Error removing suit"); return; end

    table.remove(self.tOptions.Closet, index)

    self:RebuildCloset()
    self:OnConfigShow(self.wndConfig, self.wndConfig)
end

function SuitUp:ShowSuitInEditor(suit)
    self:ClearArticlesEditor()
    self.Editor.CurrentSuit = suit
    self.Editor.CurrentArticle = nil
    if suit == nil then return; end

    local wndArticlesEditor = self.wndConfig:FindChild("SuitArticlesEditor")
    wndArticlesEditor:FindChild("SuitNameEdit"):SetText(suit.Name)
    wndArticlesEditor:FindChild("SuitNameEdit"):Enable(true)

    -- Stick the articles into the list
    local wndArticleList = wndArticlesEditor:FindChild("ArticlesList")
    for _,art in ipairs(suit.Articles) do
        local wnd = Apollo.LoadForm(self.xmlDoc, "ArticleButton", wndArticleList, self)
        wnd:SetData(art)
        if self.SuitCatalog.ArticleTypeIcons[art.ArticleType] then
            wnd:FindChild("ArticleIcon"):SetSprite(self.SuitCatalog.ArticleTypeIcons[art.ArticleType])
        end
        wnd:FindChild("ArticleType"):SetText(art.TypeName)
        wnd:FindChild("ArticleDesc"):SetText(art:Describe())
    end

    -- Add the "New article" button
    self:InsertAddArticleButton(wndArticleList)

    -- Arrange the items in the list. This *should* do them in the order we want by default.
    wndArticleList:ArrangeChildrenVert(ArrangeVert.FromTop)

    -- Scroll the article list to the top. If the previously-viewed suit had more (or longer) articles, 
    -- the pane could be scrolled down
    wndArticleList:SetVScrollPos(0) --TODO: Should this be in ClearArticlesEditor?

    --TODO: Ensure the selected Suit is visible in the suits scroll
    
    -- Show the manipulators frame for the selected suit
    local wndSuitsList = self.wndConfig:FindChild("SuitsList")
    for _,v in pairs(wndSuitsList:GetChildren()) do
        if v:GetData() == suit then
            v:SetCheck(true)
            v:FindChild("ManipulatorsFrame"):Show(true)
            v:FindChild("ManipulatorsFrame"):FindChild("DeleteSuitButton:ConfirmDeleteSuitFrame"):Show(false)
            break;
        end
    end
end

-- This sticks a button at the end of the articles. The inserted button allows you to add a new article.
function SuitUp:InsertAddArticleButton(wndArticleList)
    local wndNewArticle = Apollo.LoadForm(self.xmlDoc, "NewArticleButton", wndArticleList, self)
    local wndArticlePopupList = wndNewArticle:FindChild("NewArticlePopupList")
    wndNewArticle:AttachWindow(wndArticlePopupList)
    wndNewArticle:SetCheck(false)

    local wndArticleTypesList = wndNewArticle:FindChild("ArticleTypesList")
    for i,v in ipairs(self.SuitCatalog.ArticleDisplayOrder) do
        local atype = self.SuitCatalog.ArticlePatterns[v]
        if atype:IsEnabled() then   -- Some article types may not work in the current user's/character's configuration
            local wnd = Apollo.LoadForm(self.xmlDoc, "ArticleTypeButton", wndArticleTypesList, self)
            wnd:SetText(atype.TypeName)
            wnd:SetData(atype)
        end
    end

    wndArticleTypesList:ArrangeChildrenVert(ArrangeVert.FromTop)

end

---------------------------------------------------------------------------------------------------
function SuitUp:OnSuitNameChanging(wndHandler, wndControl, strNewText, strOldText, bAllowed)
    --TODO: Enforce some rules about the name
end

function SuitUp:OnSuitNameReturn(wndHandler, wndControl, strText)
    self.Editor.CurrentSuit.Name = strText
    
    -- Redraw the suit button in the editor
    local wndSuitsList = self.wndConfig:FindChild("SuitsList")
    for k,v in pairs(wndSuitsList:GetChildren()) do
        if v:GetData() == self.Editor.CurrentSuit then v:FindChild("SuitName"):SetText(strText); break; end
    end

    -- Redraw the suit button in the popup list
    self:RebuildCloset()

end

function SuitUp:OnSuitNameEscape(wndHandler, wndControl)
    -- Revert any changes they've made
    local wndArticlesEditor = self.wndConfig:FindChild("SuitArticlesEditor")
    wndArticlesEditor:FindChild("SuitNameEdit"):SetText(self.Editor.CurrentSuit.Name)
end

---------------------------------------------------------------------------------------------------
function SuitUp:OnAddArticleTypeButton(wndHandler, wndControl, eMouseButton)
    self.wndConfig:FindChild("NewArticleButton"):SetCheck(false)

    -- Make a new, initialized article of the proper type
    local NewArticle = self.SuitCatalog.ArticlePatterns:FromType(wndControl:GetData().ArticleType)
    
    -- Add it to the currently-being-edited suit
    table.insert(self.Editor.CurrentSuit.Articles, NewArticle)

    -- Re-show the editor
    self:ShowSuitInEditor(self.Editor.CurrentSuit)
    self:EditArticle(NewArticle)
end

---------------------------------------------------------------------------------------------------
function SuitUp:EditArticle(art)
    art = art or self.Editor.CurrentArticle
    if art == nil then return; end

    -- Make the new article active for editing
     local wndArticlesEditor = self.wndConfig:FindChild("SuitArticlesEditor")
     local wndArticleList = wndArticlesEditor:FindChild("ArticlesList")
     for k,v in pairs(wndArticleList:GetChildren()) do
        if v:GetData() == art then 
            v:SetCheck(true)
            self:OnArticleCheck(v, v, nil)
            v:GetParent():BringChildToTop(v)

            -- If the top of the article button is off screen, scroll the list
            wndArticleList:RecalculateContentExtents()
            local ButtonTop = select(2, v:GetAnchorOffsets())
            local ScrolledTop = wndArticleList:GetVScrollPos()
            local ScrolledBottom = ScrolledTop + wndArticleList:GetHeight()
            if ButtonTop >= ScrolledBottom then wndArticleList:SetVScrollPos(ButtonTop) end
            break
        end
     end
end

---------------------------------------------------------------------------------------------------
function SuitUp:OnSuitListCheck(wndHandler, wndControl, eMouseButton)
    self:ShowSuitInEditor(wndControl:GetData())
    wndControl:FindChild("ManipulatorsFrame"):Show(true)
    wndControl:FindChild("ManipulatorsFrame"):FindChild("ConfirmDeleteSuitFrame"):Show(false)
end

---------------------------------------------------------------------------------------------------
function SuitUp:OnSuitListUncheck(wndHandler, wndControl, eMouseButton)
    self:ClearArticlesEditor()
    wndControl:FindChild("ManipulatorsFrame"):Show(false)
    wndControl:FindChild("ManipulatorsFrame"):FindChild("ConfirmDeleteSuitFrame"):Show(false)
end

-- Returns the lowest point of any child of wnd
local function GetWindowFooter(wnd)
    local lowest = 0
    for k,v in pairs(wnd:GetChildren()) do
        left, top, right, bottom = v:GetRect()
        lowest = math.max(lowest, bottom)
    end
    return lowest
end

---------------------------------------------------------------------------------------------------
function SuitUp:OnArticleCheck(wndHandler, wndControl, eMouseButton)
    if wndHandler ~= wndControl then return; end -- Ignore events bubbled up from child controls like combo boxes

    self.Editor.CurrentArticle = wndControl:GetData()

    wndControl:FindChild("ArticleManipulatorsFrame"):Show(true)

    -- wndControl should be the article button. wndControl:GetData() should get the Article item
    local wndArticleEditFrame = wndControl:FindChild("EditorFrame")

    -- Hide the description, since the editor shows everything anyway and the description may change during editing
    wndControl:FindChild("ArticleDesc"):SetText(nil)

    wndControl:GetData():MakeEditor(wndArticleEditFrame)

    local n = GetWindowFooter(wndArticleEditFrame)
    local frameLeft, frameTop, frameRight, frameBottom = wndArticleEditFrame:GetAnchorOffsets()
    wndArticleEditFrame:SetAnchorOffsets(frameLeft, frameTop, frameRight, frameBottom + n)

    local buttonLeft, buttonTop, buttonRight, buttonBottom = wndControl:GetAnchorOffsets()
    wndControl:SetAnchorOffsets(buttonLeft, buttonTop, buttonRight, buttonBottom + n)

    wndControl:GetParent():ArrangeChildrenVert(ArrangeVert.FromTop)

    wndControl:EnsureChildVisible(wndControl)
end

---------------------------------------------------------------------------------------------------
function SuitUp:OnArticleUncheck(wndHandler, wndControl, eMouseButton)
    if wndHandler ~= wndControl then return; end -- Ignore events bubbled up from child controls like combo boxes

    self.Editor.CurrentArticle = nil

    wndControl:FindChild("ArticleManipulatorsFrame"):Show(false)

    -- Re-show the description, which may have changed
    wndControl:FindChild("ArticleDesc"):SetText(wndControl:GetData():Describe())

    local wndArticleEditFrame = wndControl:FindChild("EditorFrame")
    wndArticleEditFrame:DestroyChildren()

    local frameLeft, frameTop, frameRight, frameBottom = wndArticleEditFrame:GetAnchorOffsets()
    wndArticleEditFrame:SetAnchorOffsets(frameLeft, frameTop, frameRight, frameTop) -- Yes, set the top as the bottom so it has no height

    local buttonLeft, buttonTop, buttonRight, buttonBottom = wndControl:GetAnchorOffsets()
    wndControl:SetAnchorOffsets(buttonLeft, buttonTop, buttonRight, buttonBottom - frameBottom + frameTop) -- Subtract the (now old) frame height

    wndControl:GetParent():ArrangeChildrenVert(ArrangeVert.FromTop)
end


---------------------------------------------------------------------------------------------------
function SuitUp:OnDeleteArticle(wndHandler, wndControl, eMouseButton)
    -- self.Editor.CurrentSuit has the suit
    -- self.Editor.CurrentArticle is the article to delete
    if not self.tOptions.ConfirmActionDeletes then
        self:DeleteArticle(self.Editor.CurrentArticle)
        return
    end

    local wndConfirm = wndControl:FindChild("ConfirmDeleteArticleFrame")
--[[
    wndControl:BringChildToTop(wndConfirm)
    self.wndConfig:BringChildToTop(wndConfirm)
    wndControl:BringChildToTop(wndConfirm:FindChild("BG"))
    wndConfirm:ToFront()
    wndConfirm:SetFocus(true)
    wndControl:GetParent():GetParent():BringChildToTop(wndControl:GetParent())
    wndControl:GetParent():BringChildToTop(wndControl)
--]]
    wndConfirm:Show(true)
    
end

function SuitUp:OnConfirmDeleteArticle(wndHandler, wndControl, eMouseButton)
    wndControl:GetParent():Show(false) -- Not really necessary since the next line should delete the window, but in case something breaks....
    self:DeleteArticle(self.Editor.CurrentArticle)
end

function SuitUp:OnCancelDeleteArticle(wndHandler, wndControl, eMouseButton)
    wndControl:GetParent():Show(false)
end

function SuitUp:DeleteArticle(art)
    for i = 1,#self.Editor.CurrentSuit.Articles do
        if self.Editor.CurrentSuit.Articles[i] == art then
            table.remove(self.Editor.CurrentSuit.Articles, i)
            self:ShowSuitInEditor(self.Editor.CurrentSuit)
            break
        end
    end
end

---------------------------------------------------------------------------------------------------
function SuitUp:OnArticleUp(wndHandler, wndControl, eMouseButton)
    -- Okay, wndControl should be the up button. wndControl:GetParent() is the frame of manipulators,
    -- and wndControl:GetParent():GetParent() is the article button. 
    -- self.Editor.CurrentSuit is the suit being edited
    -- self.Editor.CurrentArticle should be the article being moved

    -- Get the index
    local iOrig = nil
    for i, art in ipairs(self.Editor.CurrentSuit.Articles) do
        if art == self.Editor.CurrentArticle then
            iOrig = i
           break 
        end
    end

    -- Okay, we want to move entry i up by one (to i-1)
    if iOrig == nil then Print("SuitUp: Error moving action"); return; end
    if iOrig == 1 then return; end -- Can't move the first item up

    -- Do the swap
    -- self.Editor.CurrentArticle is a strong reference to the article, so we can remove it from the Articles list and it won't gc
    table.remove(self.Editor.CurrentSuit.Articles, iOrig)
    table.insert(self.Editor.CurrentSuit.Articles, iOrig - 1, self.Editor.CurrentArticle)

    -- Re-show the articles
    local artTemp = self.Editor.CurrentArticle
    self:ShowSuitInEditor(self.Editor.CurrentSuit)
    self.Editor.CurrentArticle = artTemp
    self:EditArticle(self.Editor.CurrentArticle)
end

---------------------------------------------------------------------------------------------------
function SuitUp:OnArticleDown(wndHandler, wndControl, eMouseButton)
    -- Get the index
    local iOrig = nil
    for i, art in ipairs(self.Editor.CurrentSuit.Articles) do
        if art == self.Editor.CurrentArticle then
            iOrig = i
           break 
        end
    end

    -- Okay, we want to move entry i up by one (to i-1)
    if iOrig == nil then Print("SuitUp: Error moving action"); return; end
    if iOrig == #self.Editor.CurrentSuit.Articles then return; end -- Can't move the last item down

    -- Do the swap
    table.remove(self.Editor.CurrentSuit.Articles, iOrig)
    table.insert(self.Editor.CurrentSuit.Articles, iOrig + 1, self.Editor.CurrentArticle) 

    -- Re-show the articles
    local artTemp = self.Editor.CurrentArticle
    self:ShowSuitInEditor(self.Editor.CurrentSuit)
    self.Editor.CurrentArticle = artTemp
    self:EditArticle(self.Editor.CurrentArticle)
end

-----------------------------------------------------------------------------------------------
-- SuitUp Instance
-----------------------------------------------------------------------------------------------
local SuitUpInst = SuitUp:new()
SuitUpInst:Init()
