--[[
    LibCombo
--]]
local ComboPkg = {MAJOR = "Xowl:LibCombo-1.0", MINOR = 1}
local ArrangeVert = {FromTop = 0, FromCenter = 1, FromBottom = 2}
local TYPE = {TABLE=type({}),NUMBER=type(1),STRING=type("a"),BOOL=type(true),FUNCTION=type(function () end)}

--=============================================================================
--== Package Interface ========================================================
--=============================================================================
function ComboPkg:OnLoad()
    -- called when all dependencies are loaded
    -- if something has gone wrong, return a string with 
    -- the strError that will be passed to YOUR dependencies
end

function ComboPkg:OnDependencyError(strDep, strError)
    -- if you don't care about this dependency, return true.
    -- if you return false, or don't define this function
    -- any Addons/Packages that list you as a dependency
    -- will also receive a dependency error
    return false
end

function ComboPkg:OnSave(eType)
    -- Packages can save data, but this one doesn't
    return nil
end

function ComboPkg:OnRestore(eType, tData)
    -- Packages can restore data too
end

--=============================================================================
--== Public Interface =========================================================
--=============================================================================
--[[
    wndFrame        wnd which the combo will overlay
    tItems          Table of item entries:
                        Text            Text to display
                        AML             AML to display (instead of Text)
                        row_data        Opaque data to identify the item (not required to be unique)
                        Selectable      Boolean indicating whether the item is selectable or cosmetic
                        Tooltip         Tooltip table. If missing, the callback is used (if present)
                                        Text    Text for the tooltip
                                        AML     AML for the tooltip
    text            Initial text to display
    control_data    Opaque data identifying this XCombo
    tCallback       Table to call back for events
                        :OnXCSelect(control_data, row_data)
                        :OnXCShow(control_data)
                        :OnXCHide(control_data)
                        :OnXCItemTooltip(control_data, row_data)

    Returns a table... TBD
--]]
function ComboPkg:Create(wndFrame, tItems, text, control_data, tExternalCallbacks)
    local wndParent = wndFrame:GetParent()

    local tButton = 
    {
    }

    local tListWindow = 
    {
    }

    local wndButton = Apollo.LoadForm("XCombo.xml", "XComboButton", wndFrame:GetParent(), self)
    local wndPopout = wndButton:FindChild("XComboPopout")
    local ancl, anct, ancr, ancb = wndFrame:GetAnchorPoints()
    local offl, offt, offr, offb = wndFrame:GetAnchorOffsets()
    wndButton:SetAnchorPoints(ancl, anct, ancr, ancb)
    wndButton:SetAnchorOffsets(offl, offt, offr, offb)
    wndButton:SetData(tCallback)
    wndButton:SetText("   "..text)
    wndFrame:Show(false)
    wndButton:AttachWindow(wndPopout)
    wndButton:SetCheck(false)

    local tInternalCallbacks = 
    {
        OnSelect = function(self, wndHandler, wndControl, eMouseButton)
            local row_data = wndControl:GetData()
            wndButton:SetText("   "..row_data.Text)
            wndButton:SetCheck(false)
            if type(tExternalCallbacks.OnXCSelect) == TYPE.FUNCTION then
                tExternalCallbacks:OnXCSelect(control_data, row_data)
            end
        end,
    }

    local wndList = wndPopout:FindChild("XComboItemsList")
    for _,v in ipairs(tItems) do
        local wnd = Apollo.LoadForm("XCombo.xml", "XComboItemButton", wndList, tInternalCallbacks)
        wnd:SetData(v)
        if v.Text then wnd:SetText(v.Text) end
    end

    local nItems = #tItems
    local nTotalHeight = wndList:ArrangeChildrenVert(ArrangeVert.FromTop)
    if nItems > 0 then
        local nMaxItemsVisible = 7
        local nToShow = math.min(nMaxItemsVisible, nItems)
        local nAverageHeight = math.floor(nTotalHeight / nItems)
        local nMaxHeightVisible = nToShow * nAverageHeight
        local l, t, r, b =  wndList:GetRect()
        local nInitialHeight = b - t
        local nHeightAdjustment = nMaxHeightVisible - nInitialHeight
        self:AdjustWindowHeight(wndPopout, nHeightAdjustment)
    end

end

--=============================================================================
--== Callbacks ================================================================
--=============================================================================


--=============================================================================
--== Internals ================================================================
--=============================================================================
function ComboPkg:AdjustWindowHeight(wnd, nAdjustment)
    --TODO: Check that it's all 0-based (or at least not 1-based)
    local l, t, r, b = wnd:GetAnchorOffsets()
    wnd:SetAnchorOffsets(l, t, r, b + nAdjustment)
end

--=============================================================================
--== Instantiation ============================================================
--=============================================================================
-- Instantiate and register
Apollo.RegisterPackage(ComboPkg, ComboPkg.MAJOR, ComboPkg.MINOR, {})

-- Setting nil here (supposedly) allows this table to be collected if a newer 
-- one exists. Not sure why it doesn't go out of scope with the chunk, though.
ComboPkg = nil	
